module findmatch_Huber_B

  use helix
  use helicon
  use nchconest
  use nchelicon

  use nhelix
  use nhelicon
  use ncnhconest
  use ncnhelicon

  use verbosity_mod

  use identity_mod
  use huber_bfgs_b
  use huber_reg_mod

  implicit none

  integer, private                     :: nf,ndata
  type (filter), private               :: acorr
  real, private, dimension(:), pointer :: xcorr,space,matrix

contains

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !! Find a single matched-filter     !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  ! Minimizes residual:
  ! wght ( data * match - targt ) = 0
  subroutine findmatch(data,targt,match,niter,init,noise)
    real, dimension(:), pointer           :: data,targt
    integer                               :: niter
    logical,                     optional :: init
    real,                        optional :: noise
    type (filter) :: match
    integer       :: stat,i,nmem,nfreq
    logical       :: cginv,initchol
    real          :: sd,rnd

    call cgfindmatch(data,targt,match,niter)

  end subroutine findmatch

  !! Find match with Huber solver
  subroutine cgfindmatch(data,targt,match,niter)
    real, dimension(:), pointer           :: data,targt
    integer                               :: niter
    type (filter) :: match
    integer       :: stat

    real, dimension (:), allocatable :: l, u
    if (verb>1) write(0,*) ' Using Huber solver to solve system- there may be better way'
    
    allocate(l(size(match%flt)),u(size(match%flt)))

    l=0.
    u=0.

    call nchconest_init(data,match)   
    call lbfgs(nchconest_lop,identity_lop,huberfct,hubergdt,match%flt,niter,1e-8,0,l,u,0)

    deallocate(l,u)

  end subroutine cgfindmatch

  !! Calculates:
  !! output = input * match
  subroutine applymatch(input,output,match)

    real, dimension(:), pointer :: input, output
    type (filter)               :: match
    integer stat

    call nchelicon_init(match)
    stat=nchelicon_lop(.false.,.false.,input,output)

  end subroutine applymatch

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !! Non-stationary matched filtering !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  !! Finds non-stationary matched-filter
  subroutine findnsmatch(data,targt,match,reg,niter)
    real, dimension(:), pointer           :: data,targt
    type (nfilter)                        :: match
    type ( filter)                        :: reg
    integer                               :: mode
    integer                               :: niter

    real, dimension(:), pointer :: coeffs
    integer :: stat,nhmax,nitr,i
    real, dimension (:), allocatable :: l, u

    nhmax=size(match%hlx(1)%lag)  ! assumes all filters are same length

    allocate (coeffs(nhmax*size(match%hlx)))
    coeffs=0.

    if (verb>2) write(0,*) '  using regularized Huber solver'
    call helicon_init(reg)                  ! initialize regularizer
    call ncnhconest_init(data,match,nhmax)       ! initialize hconest

    allocate(l(nhmax*size(match%hlx)),u(nhmax*size(match%hlx)))

    l=0.
    u=0.
    call lbfgs(ncnhconest_lop,helicon_lop,huberfct,hubergdt,coeffs,niter,1.e-8,0,l,u,1)
    deallocate(l,u)

    if (verb>2) write(0,*) '  Done with solver'
    if (verb>2) write(0,*) '  Max filter value = ',maxval(coeffs)

    !! Copy coefficients into NS filter type
    do i=1,size(match%hlx)
       match%hlx(i)%flt(1:nhmax)=coeffs(1+(i-1)*nhmax:i*nhmax)
    end do

    deallocate(coeffs)

  end subroutine findnsmatch

  subroutine findnsmatch_rough(data,targt,match,reg,niter)
    real, dimension(:), pointer           :: data,targt
    type (nfilter)                        :: match
    type (nfilter)                        :: reg
    integer                               :: mode
    integer                               :: niter

    real, dimension(:), pointer :: coeffs
    integer :: stat,nhmax,nitr,i
    real, dimension (:), allocatable :: l, u
    nhmax=size(match%hlx(1)%lag)  ! assumes all filters are same length

    allocate (coeffs(nhmax*size(match%hlx)))
    coeffs=0.

    if (verb>2) write(0,*) '  using regularized Huber solver'
    call nhelicon_init(reg)                       ! initialize regularizer
    call ncnhconest_init(data,match,nhmax)       ! initialize hconest
    allocate(l(size(coeffs)),u(size(coeffs)))

    l=0.
    u=0.
    
    call lbfgs(ncnhconest_lop,nhelicon_lop,huberfct,hubergdt,coeffs,niter,1e-8,0,l,u,1)
    deallocate(l,u)

    if (verb>2) write(0,*) '  Done with solver'
    if (verb>2) write(0,*) '  Max filter value = ',maxval(coeffs)

    !! Copy coefficients into NS filter type
    do i=1,size(match%hlx)
       match%hlx(i)%flt(1:nhmax)=coeffs(1+(i-1)*nhmax:i*nhmax)
    end do

    deallocate(coeffs)

  end subroutine findnsmatch_rough

  !! Applies non-stationary matched-filter
  subroutine applynsmatch(input,output,match)
    real, dimension(:), pointer :: input, output
    type (nfilter)              :: match
    integer stat
    call ncnhelicon_init(match)
    stat=ncnhelicon_lop(.false.,.false.,input,output)
    call ncnhelicon_close()
  end subroutine applynsmatch

  !! Flags edges of cube as missing data
  !! Eliminates helical wrap around
  subroutine flagedges(missing,nwind)
    logical, dimension(:), pointer :: missing
    integer, dimension(:), pointer :: nwind
    integer :: i,j,k
    k=1
    missing=.false.
    do i=1,size(nwind)-1
       k=k*nwind(i)
       do j=1,nwind(i+1)
          missing(j*k)=.true.
       end do
    end do
  end subroutine flagedges

end module findmatch_Huber_B












